/* ── Zoom ── */

body {
  zoom: 1.2;
}

/* ── Hide TOC, swap content and nav ── */

.md-sidebar--secondary {
  display: none !important;
}

.md-main__inner {
  display: flex;
  flex-direction: row-reverse;
}

/* ── Header logo ── */

.md-header__button.md-logo img {
  width: auto;
  height: 1.8rem;
  object-fit: contain;
}

/* ── Monochrome palette ── */

/* Dark mode */
[data-md-color-scheme="slate"] {
  --md-primary-fg-color: #111111;
  --md-primary-fg-color--light: #1a1a1a;
  --md-primary-fg-color--dark: #000000;
  --md-primary-bg-color: #ffffff;
  --md-primary-bg-color--light: #ffffffb3;
  --md-accent-fg-color: #cccccc;
  --md-accent-fg-color--transparent: #cccccc33;
}

/* Light mode */
[data-md-color-scheme="default"] {
  --md-primary-fg-color: #ffffff;
  --md-primary-fg-color--light: #f5f5f5;
  --md-primary-fg-color--dark: #e0e0e0;
  --md-primary-bg-color: #111111;
  --md-primary-bg-color--light: #111111b3;
  --md-accent-fg-color: #333333;
  --md-accent-fg-color--transparent: #33333333;
}

/* ── Content links ── */

[data-md-color-scheme="slate"] .md-typeset a {
  color: #ffffff;
  text-decoration: underline;
  text-decoration-color: #555555;
  text-underline-offset: 2px;
}

[data-md-color-scheme="slate"] .md-typeset a:hover {
  text-decoration-color: #ffffff;
}

[data-md-color-scheme="default"] .md-typeset a {
  color: #000000;
  text-decoration: underline;
  text-decoration-color: #aaaaaa;
  text-underline-offset: 2px;
}

[data-md-color-scheme="default"] .md-typeset a:hover {
  text-decoration-color: #000000;
}

/* ── Headings ── */

.md-typeset h1 {
  font-weight: 700;
  letter-spacing: -0.02em;
}

.md-typeset h2 {
  font-weight: 600;
  letter-spacing: -0.01em;
}

.md-typeset h3 {
  font-weight: 600;
}

/* ── Active nav items ── */

.md-nav__link--active,
.md-nav__item .md-nav__link--active {
  font-weight: 700;
  color: var(--md-primary-bg-color) !important;
}

/* ── Sidebar section titles ── */

.md-nav--lifted > .md-nav__list > .md-nav__item--active > label.md-nav__link,
.md-nav--lifted > .md-nav__list > .md-nav__item--active > .md-nav > .md-nav__list > .md-nav__item--section > label.md-nav__link {
  text-decoration: underline;
  text-decoration-color: rgba(128, 128, 128, 0.4);
  text-underline-offset: 3px;
}

.md-nav--lifted > .md-nav__list > .md-nav__item--active > .md-nav > .md-nav__list > .md-nav__item--section {
  margin-top: 1rem;
}


/* ── Sidebar nav ── */

.md-nav__title {
  font-weight: 600;
}

.md-nav__item .md-nav__link {
  font-size: 0.72rem;
}
